
// emacs: this is -*- c++ -*-
//
//   @file    nnlo_interface.cxx
//
//            grid filling routines for interface with
//            the nnlojet code from CITATION
//
//   Copyright (C) 2015 M.Sutton (sutt@cern.ch)
//
//   $Id: nnlo_interface.h, v0.0  Thu 4 Jul 2015 10:25:18 CEST sutt $


#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <cmath>

#include <sys/stat.h>


#include "amconfig.h"

// nnlo interface header files
#include "nnlo_interface.h"
#include "nnlo_utils.h"
#include "nnlo_common.h"


#ifdef HAVE_APPL
// applgrid headers
#include "appl_grid/lumi_pdf.h"
#include "applgrid_utils.h"
#endif

#ifdef HAVE_FASTNLO
/// fastnlo headers ...
#include "fastnlo_utils.h"
#endif



// external hook functions
extern void (*appl_initptr)( const int& id, const std::string& gridname,
                             const int& nbins, const double& lo, const double& hi  );

extern void (*appl_fillptr)( const int& id, const double& obs, const double* wt );

extern void (*appl_fillrefptr)( const int& id,
                                const double& obs,
                                const double* wt );

extern void (*appl_termptr)( const int& id );

extern "C" void evolvepdf_(const double& x, const double& Q, double* xf);
extern "C" double alphaspdf_(const double& Q);


/// Declare the input and output grids
/// Declare grid-specific functions
static std::string _gridname; //  = "nnlojet.root";

// #ifdef HAVE_APPL
// //static std::int  _gridname; //  = "nnlojet.root";
// static std::vector<appl::grid*> _grids;
// static std::map<int, appl::grid*> _gridmap;
// static std::map<int, std::string> _gridnames;

// void nnlo::fill_appl_internal( appl::grid* _grid,
//                                const int& order,
//                                const double& x1, const double& x2, const double& Q2,
//                                const double& obs, const double* wgt );

// appl::grid* find( const int& id ) {
//    std::map<int, appl::grid*>::iterator gitr = _gridmap.find(id);
//    if ( gitr!=_gridmap.end() ) return gitr->second;
//    std::cerr << "find() requested grid id " << id << " not found" << std::endl;
//    return 0;
// }
// #endif


template<typename T>
std::ostream& operator<<( std::ostream& s, std::vector<T>& t ) {
  for ( size_t i=0 ; i<t.size() ; i++ ) s << "\t" << t[i];
  return s;
}


// external hook functions
extern "C" bool setup_nnlo_() {
  std::cout << "NNLOJET INFO: setting up hook functions ..." << std::endl;

  appl_initptr       = nnlo::init;
  appl_fillptr       = nnlo::fill;
  appl_fillrefptr    = nnlo::fill_ref;
  appl_termptr       = nnlo::term;

  return true;
}



/// Initialise the hook functions
static bool _ready = setup_nnlo_();



// Temporary pointer to APPLGRID lumi function for fastNLO
//const lumi_pdf *fnlopdf;



bool first_run = true;



//
// Initialisation; book interpolation grid(s)
//
void nnlo::init( const int& id0, const std::string& gridname,
                 const int& nbins, const double& lo, const double& hi ) {

   // Cheat with id to ask for fastnlo
   int id = 0;

   std::cout << banner() << std::endl;
   std::cout << "\tinit() " << date() << "\tcompiled: " << __DATE__ << std::endl;
   _gridname = gridname;
   std::cout << "### nnlo::init(): gridname is " << _gridname << std::endl;

   // From NNLOJETS common?
   //   int leading_order = inppar_.njets;
   //   int nloops        = 2;

   // Define vector with binning for FastNLO
   std::vector<double> binning;
   double binwidth = (hi-lo)/nbins;
   double flo = lo;
   for (int i=0; i<=nbins; i++) {
     binning.push_back(flo);
     flo += binwidth;
   }

   // Not sure about idea for "id"; most flexible solution would be to
   // allow an APPLGRID, fastNLO table, or both for each requested grid.
   // For the moment assume only one grid "id".
   // If id>0 --> APPLGRID, id<0 --> fastNLO, id==0 --> both.

   if ( id>=0 ) { // APPLGRID requested
#    ifdef HAVE_APPL
     init_applgrid( id, gridname, nbins, lo, hi );
#    else
     std::cerr << "### nnlo::init(): APPLGRID requested, but not linked!" << std::endl;
     std::exit(1);
#    endif
   }

   //   if ( id<=0 ) { // Test APPLGRID & fastNLO simultaneously
   if ( id<=0 ) { // fastNLO requested
#    ifdef HAVE_FASTNLO
     init_fastnlo( id, binning );
#    else
     std::cerr << "### nnlo::init(): fastNLO requested, but not linked!" << std::endl;
     std::exit(1);
#    endif
   }

   std::cout << "### nnlo::init(): End Init ..." << std::endl;

}






//
// Fill the interpolation grid with the weights
//
void nnlo::fill( const int& id, const double& obs, const double* wt ) {

   //   std::cout << "### nnlo::fill(): Fill grid ..." << std::endl;

   /// nominal kinematics
   /// use common/parfrac/ rather than common/pdf/
   //  double x1 =  pdf_.xi1;
   //  double x2 =  pdf_.xi2;
   double x1 =  parfrac_.x1;
   double x2 =  parfrac_.x2;
   double Q2 =  scales_.rscale[0];
   //  double Q2 =  Q*Q;

   if ( id>=0 ) { // APPLGRID requested
#    ifdef HAVE_APPL
     fill_applgrid( id, x1, x2, Q2, obs, wt );
#    else
     std::cerr << "### nnlo::fill(): APPLGRID requested, but not linked!" << std::endl;
     std::exit(1);
#    endif
   }

   //   if ( id<=0 ) { // Test APPLGRID & fastNLO simultaneously
   if ( id<=0 ) { // fastNLO requested
#    ifdef HAVE_FASTNLO
     fill_fastnlo( id, x1, x2, Q2, obs, wt );
#    else
     std::cerr << "### nnlo::fill(): fastNLO requested, but not linked!" << std::endl;
     std::exit(1);
#    endif
   }

   //   std::cout << "### nnlo::fill(): End filling ..." << std::endl;

}






//
// fill the applgrid reference histogram.
//
void nnlo::fill_ref( const int& id, const double& obs, const double* wt ) {

#ifdef HAVE_APPL
  fill_ref_applgrid( id, obs, wt );
#endif

#ifdef HAVE_FASTNLO
  fill_ref_fastnlo( id, obs, wt );
#endif

}



//
// Write out the grids
//
void nnlo::term( const int& id ) {

   std::cout << "nnlo::term() called " << date() << std::endl;

   std::cout << "  iterations: " << runinfo_.itmax2 << "\n"
             << "  shots:      " << runinfo_.nshota << "\n"
             << "  total:      " << runinfo_.itmax2*runinfo_.nshota
             << std::endl;

   double nweights = runinfo_.itmax2*runinfo_.nshota;

   if ( id>=0 ) { // APPLGRID requested
#    ifdef HAVE_APPL
     term_applgrid( id, nweights );
#    else
     std::cerr << "### nnlo::term(): APPLGRID requested, but not linked!" << std::endl;
     std::exit(1);
#    endif
   }

   //   if ( id<=0 ) { // Test APPLGRID & fastNLO simultaneously
   if ( id<=0 ) { // fastNLO requested
#    ifdef HAVE_FASTNLO
     term_fastnlo( id, nweights );
#    else
     std::cerr << "### nnlo::term(): fastNLO requested, but not linked!" << std::endl;
     std::exit(1);
#    endif
   }

   std::cout << "### nnlo::term(): End Term ..." << std::endl;

}
